import os
import sys
import json
import warnings
import requests
import logging
import pickle

from pyhmy import (
    cli,
    Typgpy
)

from .common import (
    log,
    msg_tag,
    harmony_dir,
    node_dir,
    node_sh_log_dir,
    bls_key_dir,
    saved_validator_path,
    imported_wallet_pass_file_dir,
    cli_bin_path,
    saved_node_config_path,
    validator_config,
    node_config,
    load_validator_config,
    load_node_config
)

if sys.version_info.major < 3:
    warnings.simplefilter("always", DeprecationWarning)
    warnings.warn(
        DeprecationWarning(
            "`AutoNode` does not support Python 2. Please use Python 3."
        )
    )
    warnings.resetwarnings()
    exit(-1)

if sys.platform.startswith('win32') or sys.platform.startswith('cygwin'):
    warnings.simplefilter("always", ImportWarning)
    warnings.warn(
        ImportWarning(
            "`AutoNode` does not work on Windows or Cygwin. Try using WSL."
        )
    )
    warnings.resetwarnings()
    exit(-1)

if sys.platform.startswith('darwin'):
    warnings.simplefilter("always", ImportWarning)
    warnings.warn(
        ImportWarning(
            "`AutoNode.node` does not work on MacOS."
        )
    )
    warnings.resetwarnings()


def _init():
    log_handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(f'{msg_tag} %(message)s')
    log_handler.setFormatter(formatter)
    logging.getLogger('AutoNode').handlers = []
    logging.getLogger('AutoNode').addHandler(log_handler)
    logging.getLogger('AutoNode').setLevel(logging.DEBUG)

    try:
        # TODO: implement logic to check for latest version of CLI and download if out of date.
        cli.environment.update(cli.download(cli_bin_path, replace=False, verbose=False))
        cli.set_binary(cli_bin_path)
    except requests.exceptions.RequestException as e:
        print(f"{Typgpy.FAIL}Request error: {e}. Exiting.{Typgpy.ENDC}", file=sys.stderr)
        raise SystemExit(e)

    try:  # Config file that should exist on setup
        load_validator_config()
    except (json.decoder.JSONDecodeError, IOError, PermissionError):
        warnings.simplefilter("once", ImportWarning)
        warnings.warn(ImportWarning("Could not import validator config, using defaults."))

    if os.path.isfile(saved_node_config_path):  # Internal file that could not exist.
        try:
            load_node_config()
        except (pickle.PickleError, IOError, PermissionError) as e:
            raise SystemExit(f"Could not import saved node config from {saved_node_config_path}, error: {e}")


_init()



import time
import json
import datetime
import traceback
import os
import logging

from pyhmy import (
    blockchain,
    cli,
    json_load,
    staking,
    Typgpy,
    exceptions
)

from .common import (
    log,
    harmony_dir,
    validator_config,
    node_config,
    check_interval,
    load_node_config,
    saved_node_config_path
)
from .validator import (
    check_and_activate,
    get_validator_information
)
from .node import (
    wait_for_node_response,
    assert_no_invalid_blocks,
)
from .util import (
    get_simple_rotating_log_handler
)
from .exceptions import (
    ResetNode
)

log_path = f"{harmony_dir}/autonode_monitor.log"
progress_check_interval = 300  # Must account for view-change
node_epoch_slack = 100  # Account for recovery time


def _check_for_hard_reset(shard_endpoint, error_ok=False):
    """
    Raises a ResetNodeError if blockchain does not match.

    Only used on testnets.
    """
    if node_config['network'] == "mainnet":
        return
    network_epoch = blockchain.get_current_epoch(endpoint=shard_endpoint)
    node_epoch = blockchain.get_current_epoch(endpoint='http://localhost:9500')
    if network_epoch == 0 or node_epoch == 0:
        return  # Don't hard reset on epoch 0, network could still be initing & resetting many times.
    else:
        node_epoch -= node_epoch_slack  # Slack to account for ops related hiccups on testnets.
    if node_epoch > network_epoch:
        log(f"{Typgpy.WARNING}Epoch of node higher than endpoint epoch, sleeping {progress_check_interval} seconds "
            f"before checking endpoint progress for hard-reset trigger.{Typgpy.ENDC}")
        time.sleep(progress_check_interval)  # check that network is not making progress
        new_network_epoch = blockchain.get_latest_header(endpoint=shard_endpoint)['epoch']
        if network_epoch < new_network_epoch < node_epoch:  # made progress so reset
            raise ResetNode(f"Blockchains don't match! Network "
                            f"epoch {new_network_epoch} < Node epoch {node_epoch + node_epoch_slack}", clean=True)
        else:
            log(f"{Typgpy.WARNING} Shard endpoint ({shard_endpoint}) is not making progress, "
                f"possible endpoint issue, or hard-reset.{Typgpy.ENDC}")
    try:
        assert_no_invalid_blocks()
    except AssertionError as e:
        raise ResetNode("INVALID BLOCK", clean=True) from e
    fb_ref_hash = blockchain.get_block_by_number(1, endpoint=shard_endpoint)['hash']
    fb_hash = blockchain.get_block_by_number(1)['hash']
    if not error_ok and fb_hash is not None and fb_ref_hash is not None and fb_hash != fb_ref_hash:
        raise ResetNode(f"Blockchains don't match! "
                        f"Block 1 hash of chain: {fb_ref_hash} != Block 1 hash of node {fb_hash}", clean=True)
    return True


def _run_monitor(shard_endpoint, duration=50):
    """
    Internal function that monitors the node for `duration` seconds.

    Hard reset (clean node.sh reset) triggers:
    1) See invalid block in logs
    2) Node's epoch is greater than network's epoch
    3) Block 1 hashes dont match (on shard)
    """
    activate_count, start_time = 0, time.time()
    while time.time() - start_time < duration:
        try:
            if node_config["auto-reset"]:
                _check_for_hard_reset(shard_endpoint)
            log(f"{Typgpy.HEADER}Validator address: {Typgpy.OKGREEN}{validator_config['validator-addr']}{Typgpy.ENDC}")
            meta_data = blockchain.get_node_metadata('http://localhost:9500/')
            log(f"{Typgpy.HEADER}Node BLS keys: {Typgpy.OKGREEN}{meta_data['blskey']}{Typgpy.ENDC}")
            log(f"{Typgpy.HEADER}Node version: {Typgpy.OKGREEN}{meta_data['version']}{Typgpy.ENDC}")
            log(f"{Typgpy.HEADER}Node network: {Typgpy.OKGREEN}{meta_data['network']}{Typgpy.ENDC}")
            log(f"{Typgpy.HEADER}Node is leader: {Typgpy.OKGREEN}{meta_data['is-leader']}{Typgpy.ENDC}")
            log(f"{Typgpy.HEADER}Node is archival: {Typgpy.OKGREEN}{meta_data['is-archival']}{Typgpy.ENDC}")
            log(f"{Typgpy.HEADER}Node shard: {Typgpy.OKGREEN}{meta_data['shard-id']}{Typgpy.ENDC}")
            log(f"{Typgpy.HEADER}Node role: {Typgpy.OKGREEN}{meta_data['role']}{Typgpy.ENDC}")
            all_val = staking.get_all_validator_addresses(endpoint=node_config['endpoint'])
            if validator_config["validator-addr"] in all_val:
                val_chain_info = get_validator_information()
                log(f"{Typgpy.HEADER}EPOS status: {Typgpy.OKGREEN}{val_chain_info['epos-status']}{Typgpy.ENDC}")
                log(f"{Typgpy.HEADER}Active status: {Typgpy.OKGREEN}{val_chain_info['active-status']}{Typgpy.ENDC}")
                log(f"{Typgpy.HEADER}Booted status: {Typgpy.OKGREEN}{val_chain_info['booted-status']}{Typgpy.ENDC}")
                log(f"{Typgpy.HEADER}Current epoch performance: {Typgpy.OKGREEN}"
                    f"{json.dumps(val_chain_info['current-epoch-performance'], indent=4)}{Typgpy.ENDC}")
                if node_config["auto-active"]:
                    if check_and_activate():
                        activate_count += 1
                    log(f"{Typgpy.HEADER}Auto activation count: {Typgpy.OKGREEN}{activate_count}{Typgpy.ENDC}")
            elif not node_config["no-validator"]:
                log(f"{Typgpy.WARNING}{validator_config['validator-addr']} is not a validator.{Typgpy.ENDC}")
            log(f"{Typgpy.HEADER}This node's latest header at {datetime.datetime.utcnow()}: "
                f"{Typgpy.OKGREEN}{json.dumps(blockchain.get_latest_headers(), indent=4)}"
                f"{Typgpy.ENDC}")
        except (exceptions.RPCError, exceptions.RequestsError, exceptions.RequestsTimeoutError) as e:
            log(f"{Typgpy.WARNING}RPC exception {e}{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}Continuing...{Typgpy.ENDC}")
        finally:
            if os.path.isfile(saved_node_config_path):
                load_node_config()
            time.sleep(check_interval)


def start(duration=float('inf')):
    """
    Start a monitor for duration seconds.

    Will throw a RestartNode exception to signal a need to restart the node.
    """
    old_logging_handlers = logging.getLogger('AutoNode').handlers.copy()
    logging.getLogger('AutoNode').addHandler(get_simple_rotating_log_handler(log_path))
    try:
        bls_keys = node_config['public-bls-keys']
        shard = json_load(cli.single_call(['hmy', 'utility', 'shard-for-bls', bls_keys[0],
                                           '--node', f'{node_config["endpoint"]}']))['shard-id']
        shard_endpoint = blockchain.get_sharding_structure(endpoint=node_config['endpoint'])[shard]['http']
        _run_monitor(shard_endpoint, duration=duration)
    except Exception as err:  # Catch all to handle recover options
        log(traceback.format_exc())
        log(f"{Typgpy.FAIL}Monitor failed with error: {err}{Typgpy.ENDC}")
        if node_config['auto-reset'] and isinstance(err, ResetNode):  # Execute Auto-Reset
            log(f"{Typgpy.WARNING}Monitor is waiting for endpoint {node_config['endpoint']} to "
                f"respond before triggering node reset.{Typgpy.ENDC}")
            wait_for_node_response(node_config['endpoint'], verbose=False, sleep=2)
            log(f"{Typgpy.WARNING}Monitor is triggering reset with clean: {node_config['clean']}{Typgpy.ENDC}")
            logging.getLogger('AutoNode').handlers = old_logging_handlers
            raise err
    logging.getLogger('AutoNode').handlers = old_logging_handlers



import os
import stat
import subprocess
import glob
import json
import time
import shutil
import logging

import requests

import pyhmy.rpc.exceptions as rpc_exception
from pyhmy import (
    blockchain,
    cli,
    Typgpy
)

from .common import (
    log,
    node_script_source,
    node_sh_log_dir,
    node_config,
    node_dir,
    bls_key_dir,
    harmony_dir
)

from .util import (
    input_with_print,
    get_simple_rotating_log_handler
)

node_sh_out_path = f"{node_sh_log_dir}/out.log"
node_sh_err_path = f"{node_sh_log_dir}/err.log"

log_path = f"{harmony_dir}/autonode_node.log"

rclone_space_buffer = 5 * 2**30  # 5GB in bytes
rclone_config = "harmony"


def _node_clean(verbose=True):
    log_dir = f"{node_dir}/latest"
    backup_log_dir = f"{node_dir}/backup_logs/{time.strftime('%Y/%m/%d/%H')}"
    os.makedirs(backup_log_dir, exist_ok=True)
    try:
        if os.path.isdir(log_dir):
            if verbose:
                log(f"{Typgpy.WARNING}[!] Moving {log_dir} to {backup_log_dir} {Typgpy.ENDC}")
            if os.path.isdir(backup_log_dir):
                shutil.rmtree(backup_log_dir)  # Should never happen, always keep latest.
            shutil.move(log_dir, backup_log_dir)
        for directory in filter(lambda e: os.path.isdir(e), os.listdir(node_dir)):
            if directory.startswith("harmony_db_") or directory.startswith(".dht"):
                path = os.path.join(node_dir, directory)
                if verbose:
                    log(f"{Typgpy.WARNING}[!] Removing {path} {Typgpy.ENDC}")
                shutil.rmtree(path)
    except shutil.Error as e:
        raise SystemExit(e)


def _rclone(rclone_sync_dir, shard, verbose=True):
    db_dir = f"{node_dir}/harmony_db_{shard}"
    node_sh_rclone_err_path = f"{node_sh_log_dir}/rclone_err_{shard}.log"
    node_sh_rclone_out_path = f"{node_sh_log_dir}/rclone_out_{shard}.log"

    try:
        # Assumption made on installed rclone config.
        rclone_path = f'harmony://pub.harmony.one/{rclone_sync_dir}/harmony_db_{shard}'
        if verbose:
            log(f"{Typgpy.WARNING}rclone harmony_db_{shard} from {rclone_path} in progress...{Typgpy.ENDC}")
        with open(node_sh_rclone_out_path, 'w') as fo:
            with open(node_sh_rclone_err_path, 'w') as fe:
                return subprocess.Popen(['rclone', 'sync', '-P', rclone_path, db_dir],
                                        env=os.environ, stdout=fo, stderr=fe)
    except OSError as e:
        if verbose:
            log(f"{Typgpy.FAIL}Failed to rclone shard {shard} db, error {e}{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}Removing shard {shard} db if it exists{Typgpy.ENDC}")
            if os.path.isdir(db_dir):
                shutil.rmtree(db_dir)


def _rclone_space_required(rclone_sync_dir, shard):
    rclone_path = f'{rclone_config}://pub.harmony.one/{rclone_sync_dir}/harmony_db_{shard}'
    try:
        space_output = subprocess.check_output(['rclone', 'size', rclone_path, '--json'], env=os.environ)
        return int(json.loads(space_output.decode('utf8'))['bytes'])
    except (subprocess.CalledProcessError, json.decoder.JSONDecodeError, KeyError) as e:
        log(f"{Typgpy.WARNING}Failed to get rclone db size requirement, error {e}{Typgpy.ENDC}")
        return 0


def _rclone_db(shard, verbose=True):
    # WARNING: rclone sync directory naming convention may change in the future
    if node_config['archival']:
        rclone_sync_dir = node_config['network'] + '.archival'
    else:
        rclone_sync_dir = node_config['network'] + '.min'

    _, _, free_space = shutil.disk_usage(os.environ['HOME'])
    free_space = free_space + rclone_space_buffer

    rclone_processes = []
    if shard == 0:
        required_space = _rclone_space_required(rclone_sync_dir, shard)
        if required_space == 0:
            log(f"{Typgpy.WARNING}Fast-sync db not available.\n"
                f"Skipping rclone shard {shard}...{Typgpy.ENDC}")
            return
        if required_space > free_space:
            log(f"{Typgpy.WARNING}[!] Insufficient disk space. Required: {required_space}, Free: {free_space} \n"
                f"Skipping fast sync...{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}[!] Suggest increasing disk space before running node!{Typgpy.ENDC}")
            return
        rclone_processes.append(_rclone(rclone_sync_dir, shard, verbose=verbose))
    else:
        required_beacon_space = _rclone_space_required(rclone_sync_dir, 0)
        required_shard_space = _rclone_space_required(rclone_sync_dir, shard)
        total_required_space = required_beacon_space + required_shard_space
        if total_required_space > free_space:
            log(f"{Typgpy.WARNING}[!] Insufficient disk space. Required: {total_required_space}, Free: {free_space} \n"
                f"Skipping fast sync...{Typgpy.ENDC}")
            log(f"{Typgpy.WARNING}[!] Suggest increasing disk space before running node!{Typgpy.ENDC}")
            return
        if required_beacon_space == 0:
            log(f"{Typgpy.WARNING}Fast-sync db not available.\n"
                f"Skipping rclone shard 0...{Typgpy.ENDC}")
        else:
            rclone_processes.append(_rclone(rclone_sync_dir, 0, verbose=verbose))
        if required_shard_space == 0:
            log(f"{Typgpy.WARNING}Fast-sync db not available.\n"
                f"Skipping rclone shard {shard}...{Typgpy.ENDC}")
        else:
            rclone_processes.append(_rclone(rclone_sync_dir, shard, verbose=verbose))

    for p in rclone_processes:
        p.wait()

    if all(p.returncode == 0 for p in rclone_processes):
        if verbose:
            log(f"{Typgpy.OKGREEN}rclone done!{Typgpy.ENDC}")
    else:
        failed_db_list = glob.glob(os.path.join(node_dir, 'harmony_db_*'))
        for db in failed_db_list:
            db_shard = os.path.basename(db).split('_')[-1]
            log(f'{Typgpy.FAIL}[!] Failed to rclone shard {db_shard} db.{Typgpy.ENDC}')
            log(f'{Typgpy.WARNING}Check {os.path.join(node_sh_log_dir, f"rclone_err_{db_shard}.log")}\n'
                f'Removing {db}.{Typgpy.ENDC}')
            shutil.rmtree(db)
        raise SystemExit('Fast sync failed.')


def _get_node_shard():
    """
    Returns node shard based on config.
    Returns None if 0 or > 1 shards are derived from the node's BLS keys.
    """
    key_shards = []
    if not node_config['public-bls-keys']:
        log(f"{Typgpy.WARNING}No saved BLS keys for node!{Typgpy.ENDC}")
    for bls_key in node_config['public-bls-keys']:
        try:
            key_shards.append(json.loads(cli.single_call(['hmy', 'utility', 'shard-for-bls', bls_key,
                                                          '--node', f'{node_config["endpoint"]}']))['shard-id'])
        except json.decoder.JSONDecodeError:
            log(f'{Typgpy.WARNING}[!] Failed to get shard for bls key {bls_key}!{Typgpy.ENDC}')
    if not key_shards:
        return None
    assert len(set(key_shards)) == 1, f"Node BLS keys can only be for 1 shard. BLS keys: {node_config['public-bls-keys']}"
    return key_shards[0]


def start(auto=False, verbose=True):
    """
    Start the harmony process and return the PID.

    Note that process is running after function return.
    """
    if subprocess.call(["pgrep", "harmony"], env=os.environ) == 0:
        raise RuntimeError("Harmony process is already running, can only start 1 node on machine with AutoNode.")
    old_logging_handlers = logging.getLogger('AutoNode').handlers.copy()
    logging.getLogger('AutoNode').addHandler(get_simple_rotating_log_handler(log_path))
    log(f"{Typgpy.HEADER}Starting node...{Typgpy.ENDC}")
    if not auto:
        log(f"{Typgpy.WARNING}You are starting a Harmony Node manually, "
            f"this is not recommended, continue? [Y]/n{Typgpy.ENDC}")
        if input_with_print("> ") not in {'Y', 'y', 'yes', 'Yes'}:
            raise SystemExit()
    os.chdir(node_dir)
    if os.path.isfile(f"{node_dir}/node.sh"):
        os.remove(f"{node_dir}/node.sh")
    try:
        r = requests.get(node_script_source, timeout=30, verify=True)
        with open("node.sh", 'w') as f:
            node_sh = r.content.decode()
            # WARNING: Hack until node.sh is changed for auto-node.
            node_sh = node_sh.replace("save_pass_file=false", 'save_pass_file=true')
            f.write(node_sh)
    except requests.exceptions.RequestException as e:
        raise SystemExit(e)
    st = os.stat("node.sh")
    os.chmod("node.sh", st.st_mode | stat.S_IEXEC)
    node_args = ["./node.sh", "-N", node_config["network"], "-z", "-f", bls_key_dir, "-M", "-S"]
    if node_config['clean'] and node_config['network'] != 'mainnet':
        if verbose:
            log(f"{Typgpy.WARNING}[!] Cleaning up old files before starting node.{Typgpy.ENDC}")
        _node_clean(verbose=verbose)
    if node_config['fast-sync']:
        if verbose:
            log(f'{Typgpy.WARNING}[!] Fast syncing before starting node.{Typgpy.ENDC}')
        shard = _get_node_shard()
        if shard is not None:
            _rclone_db(shard, verbose=verbose)
        else:
            log(f'{Typgpy.WARNING}[!] Unable to determine node shard.\n'
                f'Skipping fast sync...{Typgpy.ENDC}')
    if node_config['archival']:
        if verbose:
            log(f"{Typgpy.WARNING}[!] Starting node in archival mode.{Typgpy.ENDC}")
        node_args.append("-A")
    if node_config['no-download']:
        harmony_binary = os.path.realpath(os.path.join(node_dir, 'harmony'))
        if os.path.exists(harmony_binary):
            if 'static' in subprocess.check_output(['file', harmony_binary], env=os.environ).decode('utf8'):
                if verbose:
                    log(f"{Typgpy.WARNING}[!] Starting node with existing harmony binary.{Typgpy.ENDC}")
                node_args.append("-D")
            else:
                log(f"{Typgpy.WARNING}[!] Provided harmony binary is not statically linked. "
                    f"Path: {harmony_binary}{Typgpy.ENDC}")
                raise SystemExit(f"Harmony binary not statically linked. Path: {harmony_binary}")
        else:
            if verbose:
                log(f"{Typgpy.WARNING}[!] Starting node with released harmony binary.{Typgpy.ENDC}")
    with open(node_sh_out_path, 'w') as fo:
        with open(node_sh_err_path, 'w') as fe:
            if verbose:
                log(f"{Typgpy.HEADER}Starting node!{Typgpy.ENDC}")
            logging.getLogger('AutoNode').handlers = old_logging_handlers  # Reset logger to old handlers
            return subprocess.Popen(node_args, env=os.environ, stdout=fo, stderr=fe).pid


# TODO (low prio): create stream load printer for multiple waits_for_node_response
def wait_for_node_response(endpoint, verbose=True, tries=float("inf"), sleep=0.5):
    count = 0
    while True:
        count += 1
        try:
            blockchain.get_latest_header(endpoint=endpoint)
            break
        except (rpc_exception.RequestsError, rpc_exception.RequestsTimeoutError,
                rpc_exception.RPCError):
            if count > tries:
                raise TimeoutError(f"{endpoint} did not respond in {count} attempts (~{sleep * count} seconds)")
            if verbose and count % 10 == 0:
                log(f"{Typgpy.WARNING}Waiting for {endpoint} to respond, tried {count} times "
                    f"(~{sleep * count} seconds waited so far){Typgpy.ENDC}")
            time.sleep(sleep)
    if verbose:
        log(f"{Typgpy.HEADER}[!] {endpoint} is alive!{Typgpy.ENDC}")


def assert_no_invalid_blocks():
    if os.path.isdir(f"{node_dir}/latest"):
        files = glob.glob(f"{node_dir}/latest/zero*.log")
        if files:
            log_path = files[-1]
            assert not has_invalid_block(log_path), f"`invalid merkle root` present in {log_path}, restart AutoNode with clean option"


def is_signing(count=1500):
    """
    Read the last `count` lines and check for signing logs.
    """
    if os.path.isdir(f"{node_dir}/latest"):
        files = glob.glob(f"{node_dir}/latest/zero*.log")
        if files:
            log_path = files[-1]
            content = subprocess.check_output(["tail", "-n", count, log_path], env=os.environ).decode().split("\n")
            for line in content:
                line = line.rstrip()
                if "BINGO" in line or "HOORAY" in line:
                    return True
    return False


def has_invalid_block(log_file_path):
    assert os.path.isfile(log_file_path), f"{log_file_path} is not a file"
    try:
        with open(log_file_path, 'r', encoding='utf8') as f:
            for line in f:
                line = line.rstrip()
                if "invalid merkle root" in line:
                    return True
    except (UnicodeDecodeError, IOError):
        log(f"{Typgpy.WARNING}WARNING: failed to read `{log_file_path}` to check for invalid block{Typgpy.ENDC}")
    return False


def assert_started(timeout=60, do_log=False):
    """
    Assert the node has started within the given timeout.
    Node that rclone does NOT count towards timeout.
    """
    has_informed_rclone = False
    start_time = time.time()
    while time.time() - start_time < timeout:
        if subprocess.call(["pgrep", "rclone"], env=os.environ, stdout=subprocess.DEVNULL) == 0:
            timeout += 1
            if not has_informed_rclone and do_log:
                log(f"{Typgpy.WARNING}Fast-sync (rclone) is in progress...{Typgpy.ENDC}")
                has_informed_rclone = True
        elif subprocess.call(["pgrep", "harmony"], env=os.environ, stdout=subprocess.DEVNULL) == 0:
            if do_log:
                log(f"{Typgpy.OKGREEN}Harmony node is running...{Typgpy.ENDC}")
            return
        time.sleep(1)
    if do_log:
        log(f"{Typgpy.FAIL}Harmony node is NOT running!{Typgpy.ENDC}")
    raise AssertionError("Node failed to start")



import os
import getpass
import shutil
import time
import json
import subprocess

from pyhmy import (
    cli,
    exceptions,
    json_load,
    Typgpy,
    validator,
    staking
)

from .common import (
    log,
    validator_config,
    save_validator_config,
    node_config,
    save_node_config,
    b32_addr_len,
    bls_key_len,
    bls_key_dir,
    imported_wallet_pass_file_dir,
    cli_bin_path,
    protect_file,
    save_protected_file,
    user,
    harmony_dir,
    node_dir,
    node_sh_log_dir
)
from .util import (
    input_with_print,
    shard_for_bls
)
from .passphrase import (
    encrypt_wallet_passphrase,
    is_valid_passphrase
)


def _import_validator_address():
    if validator_config["validator-addr"] is None:
        _input_validator_address()
    if validator_config['validator-addr'] not in cli.get_accounts_keystore().values():
        log(f"{Typgpy.FAIL}Cannot create validator, {validator_config['validator-addr']} "
            f"not in shared CLI keystore.{Typgpy.ENDC}")
        raise SystemExit("Bad wallet import or validator config.")


def _input_validator_address():
    keys_list = list(sorted(cli.get_accounts_keystore().values()))
    if not keys_list:
        log(f"{Typgpy.FAIL}CLI keystore has no wallets.{Typgpy.ENDC}")
        raise SystemExit("Bad wallet import.")
    log(f"{Typgpy.HEADER}Keystore Wallet Addresses:{Typgpy.ENDC}")
    for i, addr in enumerate(keys_list):
        log(f"{Typgpy.BOLD}{Typgpy.OKBLUE}#{i}{Typgpy.ENDC}\t{Typgpy.OKGREEN}{addr}{Typgpy.ENDC}")
    log("")
    index = None
    while True:
        try:
            index = input_with_print(f"{Typgpy.HEADER}Which wallet would you like to use? {Typgpy.ENDC}\n"
                                     f"> {Typgpy.OKBLUE}#").strip()
            validator_config["validator-addr"] = keys_list[int(index)]
            log(f"{Typgpy.HEADER}Using validator {Typgpy.OKGREEN}{keys_list[int(index)]}{Typgpy.ENDC}")
            break
        except (IndexError, TypeError, ValueError) as e:
            log(f"{Typgpy.FAIL}Input `{index}` is not valid, error: {e}{Typgpy.ENDC}")


def _input_validator_field(field_name, set_func):
    existing = validator_config[field_name]
    prompt = (f"{Typgpy.HEADER}Enter {field_name}: {'' if not existing else f'({existing})'}{Typgpy.ENDC}\n"
              f"> {Typgpy.OKBLUE}")
    while True:
        raw_input = None
        try:
            raw_input = input_with_print(prompt.strip())
            if raw_input == '':
                set_func(existing)
                break
            else:
                set_func(raw_input)
                break
        except exceptions.InvalidValidatorError as e:
            if raw_input is not None:
                log(f"{Typgpy.FAIL}Input `{raw_input}` is not valid, error: {e}{Typgpy.ENDC}")


def _display_warning(field_name):
    log(f'{Typgpy.WARNING}{field_name} can not be changed after creation!{Typgpy.ENDC}')


def _bls_filter(file_name, suffix):
    if file_name.startswith('.') or not file_name.endswith(suffix):
        return False
    tok = file_name.split(".")
    if len(tok) != 2 or len(tok[0]) != bls_key_len:
        return False
    return True


def _wallet_pass_filter(file_name):
    if not file_name.startswith('one1') or not file_name.endswith(".pass"):
        return False
    tok = file_name.split(".")
    if len(tok) != 2 or len(tok[0]) != b32_addr_len:
        return False
    return True


def _save_protected_file(file_content, file_path, verbose=True):
    try:
        save_protected_file(file_content, file_path, verbose=verbose)
    except Exception as e:
        raise SystemExit(e)


def _import_bls_passphrase():
    """
    Import BLS passphrase (from user or file).
    Returns None if using imported passphrase files.

    NOTE: .pass files are SUBJECT TO CHANGE as they are files used
    by node.sh & CLI layer of the harmony program.
    """
    bls_keys = list(filter(lambda e: _bls_filter(e, '.key'), os.listdir(bls_key_dir)))
    bls_pass = list(filter(lambda e: _bls_filter(e, '.pass'), os.listdir(bls_key_dir)))
    imported_bls_keys, imported_bls_pass = set(), set()
    for k in bls_keys:
        imported_bls_keys.add(k.split('.')[0])
    for p in bls_pass:
        imported_bls_pass.add(p.split('.')[0])
    if bls_pass and not bls_keys:
        log(f"{Typgpy.WARNING}BLS passphrase file(s) were imported but no BLS key files were imported. "
            f"Passphrase files are ignored.{Typgpy.ENDC}")
        return getpass.getpass(f"Enter passphrase for {Typgpy.UNDERLINE}generated{Typgpy.ENDC} BLS key\n> ")
    if bls_keys and bls_pass:
        log(f"{Typgpy.WARNING}Importing BLS keys with BLS passphrase files (all or nothing).{Typgpy.ENDC}")
        for k in imported_bls_keys:
            if k not in imported_bls_pass:
                log(f"{Typgpy.FAIL}Imported BLS key file for {k} "
                    f"does not have an imported passphrase file.{Typgpy.ENDC}")
                raise SystemExit("Bad BLS import, missing BLS passphrase file.")
        return None
    if bls_keys:
        return getpass.getpass(f"Enter passphrase for all {Typgpy.UNDERLINE}{len(bls_keys)} "
                               f"imported{Typgpy.ENDC} BLS keys\n> ")
    return getpass.getpass(f"Enter passphrase for {Typgpy.UNDERLINE}generated{Typgpy.ENDC} BLS key\n> ")


def _import_bls(passphrase):
    """
    Import BLS keys using imported passphrase files if passphrase is None.
    Otherwise, use passphrase for imported BLS key files or generated BLS keys.

    Assumes that imported BLS key files and passphrase have been validated.
    """
    bls_keys = list(filter(lambda e: _bls_filter(e, '.key'), os.listdir(bls_key_dir)))
    if passphrase is None:  # Assumes passphrase files were imported when passphrase is None.
        if node_config['shard'] is not None:
            log(f"{Typgpy.WARNING}[!] Shard option ignored since BLS keys were imported.{Typgpy.ENDC}")
            time.sleep(3)  # Sleep so user can read message
        for k in bls_keys:
            passphrase_file = f"{bls_key_dir}/{k.replace('.key', '.pass')}"
            protect_file(passphrase_file)
            try:
                cli.single_call(['hmy', 'keys', 'recover-bls-key', f'{bls_key_dir}/{k}',
                                 '--passphrase-file', passphrase_file])
            except RuntimeError as e:
                log(f"{Typgpy.FAIL}Passphrase file for {k} is not correct. Error: {e}{Typgpy.ENDC}")
                raise SystemExit("Bad BLS import")
        return [k.replace('.key', '').replace('0x', '') for k in bls_keys]

    tmp_bls_pass_path = f"{os.environ['HOME']}/.bls_pass"
    _save_protected_file(passphrase, tmp_bls_pass_path, verbose=False)
    if len(bls_keys):
        if node_config['shard'] is not None:
            log(f"{Typgpy.WARNING}[!] Shard option ignored since BLS keys were imported.{Typgpy.ENDC}")
            time.sleep(3)  # Sleep so user can read message
        for k in bls_keys:
            try:
                cli.single_call(['hmy', 'keys', 'recover-bls-key', f'{bls_key_dir}/{k}',
                                 '--passphrase-file', tmp_bls_pass_path])
            except RuntimeError as e:
                log(f"{Typgpy.FAIL}Passphrase for {k} is not correct. Error: {e}{Typgpy.ENDC}")
                raise SystemExit("Bad BLS import")
            _save_protected_file(passphrase, f"{bls_key_dir}/{k.replace('.key', '.pass')}")
        os.remove(tmp_bls_pass_path)
        return [k.replace('.key', '').replace('0x', '') for k in bls_keys]
    elif node_config['shard'] is not None:
        assert isinstance(node_config['shard'], int), f"shard: {node_config['shard']} is not an integer."
        while True:
            key = json_load(cli.single_call(['hmy', 'keys', 'generate-bls-key', '--passphrase-file', tmp_bls_pass_path]))
            public_bls_key, bls_file_path = key['public-key'], key['encrypted-private-key-path']
            shard_id = json_load(cli.single_call(['hmy', '--node', f'{node_config["endpoint"]}', 'utility',
                                                  'shard-for-bls', public_bls_key]))['shard-id']
            if int(shard_id) != node_config['shard']:
                os.remove(bls_file_path)
            else:
                log(f"{Typgpy.OKGREEN}Generated BLS key for shard {shard_id}: "
                    f"{Typgpy.OKBLUE}{public_bls_key}{Typgpy.ENDC}")
                break
        shutil.move(bls_file_path, bls_key_dir)
        _save_protected_file(passphrase, f"{bls_key_dir}/{key['public-key'].replace('0x', '')}.pass")
        os.remove(tmp_bls_pass_path)
        return [public_bls_key]
    else:
        key = json_load(cli.single_call(['hmy', 'keys', 'generate-bls-key', '--passphrase-file', tmp_bls_pass_path]))
        public_bls_key = key['public-key']
        bls_file_path = key['encrypted-private-key-path']
        shard_id = json_load(cli.single_call(['hmy', '--node', f'{node_config["endpoint"]}', 'utility',
                                              'shard-for-bls', public_bls_key]))['shard-id']
        log(f"{Typgpy.OKGREEN}Generated BLS key for shard {shard_id}: {Typgpy.OKBLUE}{public_bls_key}{Typgpy.ENDC}")
        shutil.move(bls_file_path, bls_key_dir)
        _save_protected_file(passphrase, f"{bls_key_dir}/{key['public-key'].replace('0x', '')}.pass")
        os.remove(tmp_bls_pass_path)
        return [public_bls_key]


def _assert_same_shard_bls_keys(public_keys):
    ref_shard = None
    for key in public_keys:
        shard = shard_for_bls(key)
        if ref_shard is None:
            ref_shard = shard
        assert shard == ref_shard, f"Bls keys {public_keys} are not for same shard, {shard} != {ref_shard}"


def make_directories():
    os.makedirs(harmony_dir, exist_ok=True)
    os.makedirs(node_dir, exist_ok=True)
    os.makedirs(node_sh_log_dir, exist_ok=True)
    os.makedirs(bls_key_dir, exist_ok=True)
    os.makedirs(imported_wallet_pass_file_dir, exist_ok=True)


def update_cli():
    cli.download(cli_bin_path, replace=True)


def setup_node_config():
    """
    Setup configs needed to run a node.
    """
    make_directories()
    bls_passphrase = _import_bls_passphrase()
    public_bls_keys = _import_bls(bls_passphrase)
    _assert_same_shard_bls_keys(public_bls_keys)
    node_config['public-bls-keys'] = public_bls_keys
    save_node_config()


def setup_validator_config():
    """
    Setup configs needed to handle a validator.
    """
    if node_config['no-validator']:
        log(f"{Typgpy.WARNING}Node config specify no validator automation, skipping setup...{Typgpy.ENDC}")
        return
    assert node_config['public-bls-keys'], f"sanity check: node config is not setup."

    make_directories()
    interactive_setup_validator()
    shard_id = json_load(cli.single_call(['hmy', '--node', f'{node_config["endpoint"]}', 'utility',
                                          'shard-for-bls', node_config['public-bls-keys'][0]]))['shard-id']
    log("~" * 110)
    log(f"Shard ID: {shard_id}")
    log(f"Saved Validator Information: {json.dumps(validator_config, indent=4)}")
    save_validator_config()
    log("~" * 110)


def setup_wallet_passphrase():
    """
    Setup wallet passphrase (encrypted).
    """
    if node_config['no-validator']:
        log(f"{Typgpy.WARNING}Node config specify no validator automation, skipping wallet setup...{Typgpy.ENDC}")
        return
    passphrase = get_wallet_passphrase()
    log(f"{Typgpy.HEADER}Encrypting and saving wallet passphrase.{Typgpy.ENDC}")
    save_wallet_passphrase(passphrase)


def interactive_setup_validator():
    """
    Interactively take validator information, always ask for all fields
    """
    if not validator_config['validator-addr']:
        _input_validator_address()

    v = validator.Validator(validator_config['validator-addr'])
    if validator_config['validator-addr'] in staking.get_all_validator_addresses(endpoint=node_config['endpoint']):
        v.load_from_blockchain(node_config['endpoint'])
        # Can immediately load validator config since information is from on-chain data.
        for key, value in v.export().items():
            assert value is not None, f"sanity check: validated config ({key}) should not be None"
            validator_config[key] = str(value)

    update_validator = True
    if all(el is not None for el in validator_config.values()):
        log(f"{Typgpy.HEADER}Validator Config:{Typgpy.OKGREEN} {json.dumps(validator_config, indent=2)}{Typgpy.ENDC}")
        if input_with_print("Update validator? [Y]/n\n> ").lower() not in {'y', 'yes'}:
            try:
                v.load(validator_config)  # The act of loading will check params
                update_validator = False
            except exceptions.InvalidValidatorError as e:
                log(f"{Typgpy.FAIL}Failed to load validator information, error {e}{Typgpy.ENDC}")
                log(f"{Typgpy.WARNING}Re-enter validator info:{Typgpy.ENDC}")
                update_validator = True

    if update_validator:
        _input_validator_field('name', v.set_name)
        _input_validator_field('website', v.set_website)
        _input_validator_field('security-contact', v.set_security_contact)
        _input_validator_field('identity', v.set_identity)
        _input_validator_field('details', v.set_details)
        _input_validator_field('min-self-delegation', v.set_min_self_delegation)
        _input_validator_field('max-total-delegation', v.set_max_total_delegation)
        _input_validator_field('amount', v.set_amount)
        _display_warning('max-rate')
        _input_validator_field('max-rate', v.set_max_rate)
        _display_warning('max-change-rate')
        _input_validator_field('max-change-rate', v.set_max_change_rate)
        _input_validator_field('rate', v.set_rate)
        for key, value in v.export().items():
            assert value is not None, f"sanity check: validated config ({key}) should not be None"
            validator_config[key] = str(value)


def get_wallet_passphrase():
    """
    Get wallet passphrase from wallet passphrase directory, if present.
    Otherwise get it interactively from user.
    """
    wallet_pass = filter(_wallet_pass_filter, os.listdir(imported_wallet_pass_file_dir))
    for p in wallet_pass:
        if validator_config['validator-addr'] == p.split('.')[0]:
            passphrase_file = f"{imported_wallet_pass_file_dir}/{p}"
            try:
                with open(passphrase_file, 'r', encoding='utf8') as f:
                    return f.read().strip()
            except (IOError, PermissionError) as e:
                raise AssertionError(f"Failed to import passphrase from {passphrase_file}, error: {e}")
    return getpass.getpass(f"Enter wallet passphrase for {validator_config['validator-addr']}\n> ")


def save_wallet_passphrase(passphrase):
    """
    Encrypt and save wallet passphrase in node config.
    """
    is_node_running = subprocess.call(["pgrep", "harmony"], env=os.environ, stdout=subprocess.DEVNULL) == 0
    assert is_node_running, "Harmony process is not running, cannot save wallet passphrase"
    addr = validator_config["validator-addr"]
    assert addr, "Validator was not setup, cannot save passphrase"
    assert is_valid_passphrase(passphrase, addr), f"Invalid passphrase for {addr}"
    node_config["encrypted-wallet-passphrase"] = encrypt_wallet_passphrase(passphrase)
    save_node_config()



import gzip
import os
import logging
import logging.handlers
import sys

import pexpect
from pyhmy import cli
from pyhmy import (
    Typgpy,
    json_load
)

from .common import (
    log,
    node_config,
    validator_config,
    msg_tag
)
from .passphrase import (
    decrypt_wallet_passphrase,
    is_valid_passphrase
)
from .exceptions import (
    InvalidWalletPassphrase
)


class _GZipRotator:
    """A simple zip rotator for logging"""

    def __call__(self, source, dest):
        os.rename(source, dest)
        f_in = open(dest, 'rb')
        f_out = gzip.open("%s.gz" % dest, 'wb')
        f_out.writelines(f_in)
        f_out.close()
        f_in.close()
        os.remove(dest)


def get_wallet_passphrase():
    """
    Gets encrypted passphrase from node_config, unencrypt, validate, and return.

    Raises InvalidWalletPassphrase if encrypted passphrase is invalid.
    """
    passphrase = decrypt_wallet_passphrase(node_config["encrypted-wallet-passphrase"])
    if not is_valid_passphrase(passphrase, validator_config["validator-addr"]):
        raise InvalidWalletPassphrase()
    return passphrase


def pexpect_input_wallet_passphrase(proc, passphrase, prompt="Enter wallet keystore passphrase:\r\n"):
    """
    Interactively input the wallet passphrase to the given pexpect child process
    """
    assert isinstance(proc, pexpect.pty_spawn.spawn)
    proc.expect(prompt)
    proc.sendline(passphrase)


def pexpect_input_wallet_creation_passphrase(proc, passphrase):
    """
    This will enter the `passphrase` interactively given the pexpect child program, `proc`.
    """
    assert isinstance(proc, pexpect.pty_spawn.spawn)
    pexpect_input_wallet_passphrase(proc, passphrase, prompt="Enter passphrase:\r\n")
    pexpect_input_wallet_passphrase(proc, passphrase, prompt="Repeat the passphrase:\r\n")
    proc.expect("\n")


def input_with_print(prompt_str, auto_interaction=None):
    if auto_interaction is not None:
        return auto_interaction
    user_input = input(prompt_str).strip()
    sys.stdout.write(f"{Typgpy.ENDC}")
    sys.stdout.flush()
    log(f"{Typgpy.OKBLUE}You entered: `{Typgpy.OKGREEN}{user_input}{Typgpy.OKBLUE}`{Typgpy.ENDC}")
    return user_input


def check_min_bal_on_s0(address, amount, endpoint=node_config['endpoint'], timeout=30):
    balances = json_load(cli.single_call(['hmy', '--node', endpoint, 'balances', address], timeout=timeout))
    for bal in balances:
        if bal['shard'] == 0:
            return bal['amount'] >= amount


def get_simple_rotating_log_handler(log_file_path, max_size=5 * 1024 * 1024):
    """
    A simple log handler with no level support.
    Used purely for the output rotation.

    `max_size` of log file is in bytes.
    """
    log_formatter = logging.Formatter(f'{msg_tag} %(message)s')
    handler = logging.handlers.RotatingFileHandler(log_file_path, mode='a', maxBytes=max_size,
                                                   backupCount=5, encoding=None, delay=0)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(log_formatter)
    handler.rotator = _GZipRotator()
    return handler


def shard_for_bls(public_bls_key):
    """
    Fetch the shard for the BLS key.

    NOTE: Expect to be invalid one we have resharding.
    """
    return json_load(cli.single_call(['hmy', '--node', f'{node_config["endpoint"]}', 'utility',
                                      'shard-for-bls', public_bls_key]))['shard-id']



import subprocess
import os
import base64

import pexpect
from cryptography.fernet import Fernet
from cryptography.fernet import InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from pyhmy import (
    cli,
    Typgpy
)

from .common import (
    validator_config,
    node_config,
    log
)
from .exceptions import (
    InvalidWalletPassphrase
)


def _get_harmony_pid():
    try:
        return subprocess.check_output(["pgrep", "harmony"], env=os.environ).strip()
    except (subprocess.TimeoutExpired, subprocess.CalledProcessError):
        return b'0'


def _get_process_info(pid):
    assert isinstance(pid, bytes)
    pid = pid.strip()
    try:
        if int(pid) <= 0:
            return b'0'
        proc_start = subprocess.check_output(["ps", "-p", pid.decode(), "-o", "lstart="], env=os.environ).strip()
        proc_command = subprocess.check_output(["ps", "-p", pid.decode(), "-o", "command="], env=os.environ).strip()
        return proc_start + b' ' + proc_command
    except (subprocess.TimeoutExpired, subprocess.CalledProcessError, ValueError):
        return b'0'


def _get_node_based_salt():
    key = ''.join([str(x) for x in node_config["public-bls-keys"]]
                  + [str(validator_config["validator-addr"]), str(validator_config["identity"])]).encode()
    return hmac.HMAC(key, hashes.SHA256(), backend=default_backend()).finalize()


def _derive_wallet_encryption_key():
    """
    Create the wallet encryption key based on:
        PBKDF2HMAC(PID(harmony) + ProcessInfo(harmony), salt=HMAC(node_bls_public_keys, Validator_Addr, Validator_ID))

    This means that the encryption key is only valid for when AutoNode has a harmony node process running.
    """
    pid = _get_harmony_pid().strip()
    proc_info = _get_process_info(pid).strip()
    salt = _get_node_based_salt()
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=10000,
        backend=default_backend()
    )
    data = pid + b' ' + proc_info
    return base64.urlsafe_b64encode(kdf.derive(data))


def encrypt_wallet_passphrase(passphrase):
    """
    Encrypt the given passphrase.

    Returned string can be stored on disk, will be invalidated once harmony process is stopped.
    """
    assert isinstance(passphrase, str)
    return Fernet(_derive_wallet_encryption_key()).encrypt(passphrase.encode())


def decrypt_wallet_passphrase(encrypted_wallet_passphrase):
    """
    Decrypt the given encrypted passphrase.
    """
    assert isinstance(encrypted_wallet_passphrase, bytes)
    try:
        return Fernet(_derive_wallet_encryption_key()).decrypt(encrypted_wallet_passphrase).decode()
    except InvalidToken:
        raise InvalidWalletPassphrase()


def is_valid_passphrase(passphrase, validator_address):
    """
    Validate the given passphrase, can be an expensive call.
    """
    cmd = ["hmy", "keys", "check-passphrase", validator_address]
    try:
        proc = cli.expect_call(cmd)
        proc.expect("Enter wallet keystore passphrase:\r\n")
        proc.sendline(passphrase)
        proc.expect("Valid passphrase\r\n")
        proc.expect(pexpect.EOF)
        return True
    except RuntimeError as e:
        log(f"{Typgpy.FAIL}Failed to verify passphrase due to: {e}{Typgpy.ENDC}")
        return False
    except pexpect.ExceptionPexpect as e:
        log(f"{Typgpy.FAIL}Failed to verify passphrase due to:\n{e.get_trace()}{Typgpy.ENDC}")
        return False



import subprocess
from setuptools import setup

setup(
    name='AutoNode',
    version=f"0.6.4",
    description="AutoNode Python Library",
    author='Daniel Van Der Maden',
    author_email='daniel@harmony.one',
    url="http://harmony.one/auto-node",
    packages=['AutoNode'],
    install_requires=[
        'requests==2.23.0',
        'pexpect==4.8.0',
        'cryptography==2.9.2',
        'pyhmy',
    ]
)



"""
Library of tools to be used by the AutoNode Daemon.
"""
import os
import subprocess
import time

from .common import (
    saved_node_config_path,
    save_node_config,
    saved_validator_path,
    bls_key_dir,
    node_config,
)
from .node import (
    start as start_node
)
from .validator import (
    setup as setup_validator,
    assert_node_started
)
from .monitor import (
    start as start_monitor,
    ResetNode
)
from .initialize import (
    save_wallet_passphrase
)
from .util import (
    get_wallet_passphrase,
)
from .exceptions import (
    InvalidWalletPassphrase
)

name = f"autonoded"
services = [
    "monitor",
    "node",
]


def _validate_config(for_node=False):
    """
    Validates the given config.
    Should be used prior to running a major AutoNode operation.
    """
    required_files = [
        saved_node_config_path,
    ]
    if not for_node:
        required_files.append(saved_validator_path)
    if any(not os.path.isfile(p) for p in required_files):
        raise SystemExit(f"AutoNode was not initialized properly. "
                         f"One or more files are missing: {required_files}")
    if for_node:
        # Assumes that .pass files are always present -- artifact of node.sh, should be amended in the future
        files_in_bls_dir = set(os.listdir(bls_key_dir))
        for bls_key in node_config['public-bls-keys']:
            key_file, pass_file = f"{bls_key}.key", f"{bls_key}.pass"
            if key_file not in files_in_bls_dir:
                raise SystemExit(f"{bls_key} in node config, but {key_file} not found in "
                                 f"BLS key directory at {bls_key_dir}")
            if pass_file not in files_in_bls_dir:
                raise SystemExit(f"{bls_key} in node config, but {pass_file} not found in "
                                 f"BLS key directory at {bls_key_dir}")


def run_node(hard_reset_recovery=False, duration=float('inf')):
    """
    Main function to run a harmony node.
    Will block for the `duration`.
    """
    print(f"Running node for {duration} seconds. Hard reset: {hard_reset_recovery}")
    start_time = time.time()
    _validate_config(for_node=True)
    pid = None
    try:
        pid = start_node(auto=True, verbose=True)
        if hard_reset_recovery:
            setup_validator(hard_reset_recovery=True)
        while time.time() - start_time < duration:
            time.sleep(1)
    finally:
        if pid is not None:
            print(f"Killing harmony process, pid: {pid}")
            subprocess.check_call(["kill", "-2", str(pid)], env=os.environ)


def _reset_node(recover_service_name, error):
    """
    Internal function to reset a node during hard reset.
    """
    assert isinstance(error, ResetNode)
    assert recover_service_name in services, f"{recover_service_name} is not a valid service."

    print(f"Resetting Node: {error}")
    if node_config['network'] == 'mainnet':
        print("WARNING: cannot reset mainnet node, ignoring...")
        return

    passphrase = get_wallet_passphrase()

    # Set flags to indicate that node is in hard-reset recovery mode.
    node_config['_is_recovering'] = True
    save_node_config()

    for service in filter(lambda e: e.startswith("node"), services):
        daemon_name = f"{name}@{service}.service"
        command = ["systemctl", "--user", "stop", daemon_name]
        print(f"Stopping daemon {daemon_name}")
        try:
            subprocess.check_call(command, env=os.environ)
        except subprocess.CalledProcessError as e:
            print(f"Unable to stop service '{daemon_name}'")
            raise e

    subprocess.call(["killall", "harmony"], env=os.environ)  # OK if this fails, so use subprocess.call
    time.sleep(5)  # wait for node shutdown

    daemon_name = f"{name}@{recover_service_name}.service"
    command = ["systemctl", "--user", "start", daemon_name]
    print(f"Starting daemon {daemon_name}")
    try:
        subprocess.check_call(command, env=os.environ)
    except subprocess.CalledProcessError as e:
        print(f"Unable to start service '{daemon_name}'")
        raise e

    try:
        assert_node_started()
        save_wallet_passphrase(passphrase)
    except (AssertionError, InvalidWalletPassphrase) as e:
        print(f"Could not re-auth wallet, error {e}")
        print(f"Continuing...")
    finally:
        # Set flags to indicate that node finished with hard-reset recovery.
        node_config['_is_recovering'] = False
        save_node_config()


def run_monitor(duration=float('inf')):
    """
    Main function to run the monitor.
    """
    print(f"Running monitor for {duration} seconds.")
    _validate_config(for_node=False)
    while True:
        try:
            # Monitor will raise a ResetNode exception to trigger a node reset, otherwise it will gracefully exit.
            start_monitor(duration=duration)
            break
        except ResetNode as error:
            if node_config['auto-reset']:
                _reset_node("node", error)
    print("Terminating monitor.")
